/*
 * Formula.hpp
 *
 *  Created on: Jul 21, 2010
 *      Author: crueger
 */

#ifndef FORMULA_HPP_
#define FORMULA_HPP_

#include <vector>
#include <iosfwd>
#include <iterator>

#include "Exceptions/ParseError.hpp"

#include "types.hpp"

class element;

class Formula
{
public:
  typedef element*                                      key_type;
  typedef unsigned int                                  mapped_type;
  typedef std::pair<key_type, mapped_type>              value_type;
  typedef std::vector<unsigned int>                     set_t;

  template <class result_type>
  class _iterator : public std::iterator <std::bidirectional_iterator_tag,
                                         result_type>
  {
  public:
    _iterator(set_t&);
    _iterator(set_t&,size_t);
    _iterator(const _iterator &rhs);
    ~_iterator();
    _iterator   &operator=(const _iterator &);
    bool         operator==(const _iterator&);
    bool         operator!=(const _iterator&);
    _iterator    operator++();
    _iterator    operator++(int);
    _iterator    operator--();
    _iterator    operator--(int);
    result_type  operator*();
    result_type *operator->();
  private:
    set_t *set;
    size_t pos; // cannot use an iterator, because we need a position to access the element
  };

  typedef _iterator<value_type>                 iterator;
  typedef _iterator<const value_type>           const_iterator;
  typedef std::reverse_iterator<iterator>       reverse_iterator;
  typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

  Formula();
  Formula(const std::string&);
  Formula(const Formula&);
  virtual ~Formula();

  Formula &operator=(const Formula&);

  std::string toString() const;
  void fromString(const std::string&) throw(ParseError);
  bool checkOut(std::ostream*) const;

  unsigned int getElementCount() const;
  bool hasElement(const element*)       const;
  bool hasElement(atomicNumber_t)       const;
  bool hasElement(const std::string&)   const;

  void operator+=(const element*);
  void operator+=(atomicNumber_t);
  void operator+=(const std::string&);

  void operator-=(const element*);
  void operator-=(atomicNumber_t);
  void operator-=(const std::string&);

  void addElements(const element*,unsigned int);
  void addElements(atomicNumber_t,unsigned int);
  void addElements(const std::string&,unsigned int);

  // only const versions, because someone might try to increment a previously
  // not set element
  const unsigned int operator[](const element*) const;
  const unsigned int operator[](atomicNumber_t) const;
  const unsigned int operator[](std::string)    const;

  bool operator==(const Formula&) const;
  bool operator!=(const Formula&) const;

  iterator       begin();
  const_iterator begin() const;
  iterator       end();
  const_iterator end()   const;

  reverse_iterator       rbegin();
  const_reverse_iterator rbegin() const;
  reverse_iterator       rend();
  const_reverse_iterator rend()   const;

  void clear();

private:
  mutable set_t elementCounts; // we might need to resize even at const places
  unsigned int numElements;
};

std::ostream &operator<<(std::ostream&,const Formula&);

#endif /* FORMULA_HPP_ */
